@@ -1,24 +1,29 @@
-import logging, json, datetime, pickle,re
+import logging, json, datetime, pickle, re
 from json.decoder import JSONDecodeError
 import requests
 from requests.exceptions import ConnectionError
 import urllib3
 from dataclasses import dataclass,field
-from absbox.local.util import mkTag,isDate,flat
-from absbox.local.component import mkPool,mkAssumption,mkAssumption2
+from absbox.local.util import mkTag, isDate, flat, guess_pool_locale, mapValsBy
+from absbox.local.component import mkPool, mkAssumption, mkAssumption2, mkPricingAssump
 from absbox.local.base import *
 import pandas as pd
-from pyspecter import S,query
+from pyspecter import query
 
+from importlib.metadata import version
+
+VERSION_NUM = version("absbox")
 urllib3.disable_warnings()
 
+
 @dataclass
 class API:
-    url:str
-    lang:str = "chinese"
+    url: str
+    lang: str = "chinese"
     server_info = {}
-    version = "0","11","1"
-    hdrs = {'Content-type': 'application/json', 'Accept': 'text/plain'}
+    #version = "0","14","1"
+    version = VERSION_NUM.split(".")
+    hdrs = {'Content-type': 'application/json', 'Accept': 'text/plain','Accept':'*/*'}
 
     def __post_init__(self):
         try:
@@ -30,75 +35,85 @@ class API:
 
         echo = json.loads(_r)
         self.server_info = echo
-        x,y,z = echo['version'].split(".")
-        logging.info(f"Connect with engine version {echo['version']} successfully")
+        x,y,z = echo['_version'].split(".")
+        logging.info(f"Connect with engine {self.url} version {echo['_version']} successfully")
         if self.version[1] != y:
             logging.error(f"Failed to init the api instance, lib support={self.version} but server version={echo['version']} , pls upgrade your api package by: pip -U absbox")
             return
 
-    def build_req(self, deal, assumptions=None, pricing=None, read=None) -> str:
+    def build_req(self, deal, assumptions=None, pricing=None) -> str:
+        r = None
         _assump = None 
+        _deal = deal.json
+        _pricing = mkPricingAssump(pricing) if pricing else None
         if isinstance(assumptions, dict):
-            _assump = mkTag(("Multiple", { scenarioName:mkAssumption2(a) for (scenarioName,a) in assumptions.items()}))
-        elif isinstance(assumptions, list) or  isinstance(assumptions, tuple):   
-            _assump = mkTag(("Single",mkAssumption2(assumptions)))
+            # _assump = { scenarioName:mkAssumption2(a) for (scenarioName,a) in assumptions.items()}
+            _assump = mapValsBy(assumptions, mkAssumption2)
+            r = mkTag(("MultiScenarioRunReq",[_deal, _assump, _pricing]))
+        elif isinstance(assumptions, list) :   
+            # _assump = mkTag(("Single",mkAssumption2(assumptions)))
+            _assump = mkAssumption2(assumptions)
+            r = mkTag(("SingleRunReq",[_deal, _assump, _pricing]))
+        elif assumptions is None:
+            r = mkTag(("SingleRunReq",[_deal, None, _pricing]))
+        return json.dumps(r , ensure_ascii=False)
+
+    def build_pool_req(self, pool, assumptions, read=None) -> str:
+        r = None
+        if isinstance(assumptions, list):
+            r = mkTag(("SingleRunPoolReq"
+                      ,[mkPool(pool)
+                        ,mkAssumption2(assumptions)])) 
+        elif isinstance(assumptions, dict):
+            r = mkTag(("MultiScenarioRunPoolReq"
+                      ,[mkPool(pool)
+                        ,mapValsBy(assumptions, mkAssumption2)])) 
         else:
-            _assump = None
-        return json.dumps({"deal": deal.json
-                          ,"assump": _assump
-                          ,"bondPricing": deal.read_pricing(pricing) if (pricing is not None) else None}
-                          , ensure_ascii=False)
+            raise RuntimeError("Error in build pool req")
+        return json.dumps(r , ensure_ascii=False)
 
-    def build_pool_req(self, pool, assumptions=[], read=None) -> str:
-        return json.dumps({"pool": mkPool(pool)
-                          ,"pAssump": mkAssumption2(assumptions)}
-                          ,ensure_ascii=False)
-
-    def _validate_assump(self,x,e,w):
-        def asset_check(_e,_w):
-            return _e,_w
-        a = x['assump']
-        asset_ids = set(range(len(query(x,['deal','contents','pool','assets']))))
-        match a:
-            case {'tag':'Single','contents':{'tag':'PoolLevel'}}:
-                return [True,e,w]
-            case {'tag':'Multiple','contents':{'tag':'PoolLevel'}}:
-                return [True,e,w]
-            case {'tag':'Single', 'contents':{'tag':'ByIndex', 'contents':(assumps,_)}}:
-                _ids = set(flat([ assump[0] for assump in assumps ]))
-                if not _ids.issubset(asset_ids):
-                    e.append(f"Not Valid Asset ID:{_ids - asset_ids}")
-                missing_asset_id = asset_ids - _ids
-                if len(missing_asset_id)>0:
-                    w.append(f"Missing Asset to set assumption:{missing_asset_id}")            
-            case {'tag':'Multiple', 'contents':scenarioMap}:
-                for k,v in scenarioMap.items():
-                    if v['tag']=='PoolLevel':
-                        continue
-                    _ids = set(flat([ _a[0] for _a in v['contents'][0]]))
+    def _validate_assump(self, x, e, w):
+        def asset_check(_e, _w):
+            return _e, _w
+        def _validate_single_assump(z):
+            match z:
+                case {'tag': 'PoolLevel'}:
+                    return [True, e, w]
+                case {'tag':'ByIndex', 'contents':(assumps, _)}:
+                    _ids = set(flat([ assump[0] for assump in assumps ]))
                     if not _ids.issubset(asset_ids):
-                        e.append(f"Scenario:{k},Not Valid Asset ID:{_ids - asset_ids}")
-                    missing_asset_id = asset_ids - _ids
-                    if len(missing_asset_id)>0:
-                        w.append(f"Scenario:{k},Missing Asset to set assumption:{missing_asset_id}")
-            case None:
-                return [True,e,w]
-            case _ :
-                raise RuntimeError(f"Failed to match:{a}")
-        if len(e)>0:
-            return [False,e,w]
-        return [True,e,w]
+                        e.append(f"Not Valid Asset ID:{_ids - asset_ids}")
+                    if len(missing_asset_id := asset_ids - _ids) > 0:
+                        w.append(f"Missing Asset to set assumption:{missing_asset_id}")            
+                case None:
+                    return [True, e, w]
+                case _:
+                    raise RuntimeError(f"Failed to match:{a}")
+        a = x['contents'][1]
+        asset_ids = set(range(len(query(x['contents'][0], ['contents', 'pool', 'assets']))))
+        match x:
+            case {"tag":"SingleRunReq","contents":[d,ma,_]}:
+                _validate_single_assump(ma)
+            case {"tag":"MultiScenarioRunReq","contents":[d,mam,_]}:
+                mapValsBy(mam, _validate_single_assump)
+            case {"tag":"MultiDealRunReq","contents":[dm,ma,_]}:
+                _validate_single_assump(ma)
+
+        if len(e) > 0:
+            return [False, e, w]
+        return [True, e, w]
 
     def validate(self, _r) -> list:
         error = []
         warning = []
         _r = json.loads(_r)
-        __d = _r['deal']
+        __d = _r['contents'][0]
         _d = __d['contents']
         valid_acc = set(_d['accounts'].keys())
         valid_bnd = set(_d['bonds'].keys())
         valid_fee = set(_d['fees'].keys())
         _w = _d['waterfall']
+        #print("Waterfall",_w)
 
         _,error,warning = self._validate_assump(_r,error,warning)
 
@@ -149,7 +164,11 @@ class API:
                     case 'PayFeeResidual':
                         if (action['contents'][1] not in valid_acc) \
                             or (action['contents'][2] not in valid_fee):
-                            error.append(f"{wn},{idx}")        
+                            error.append(f"{wn},{idx}")
+                    case 'PayFeeResidual':
+                        if (action['contents'][1] not in valid_acc) \
+                            or (action['contents'][2] not in valid_fee):
+                            error.append(f"{wn},{idx}")
 
         if warning:
             logging.warning(f"Warning in modelling:{warning}")
@@ -166,7 +185,6 @@ class API:
             deal,
             assumptions=None,
             pricing=None,
-            custom_endpoint=None,
             read=True,
             position=None,
             timing=False):
@@ -175,15 +193,8 @@ class API:
             assumptions = pickle.load(assumptions)
 
         # if run req is a multi-scenario run
-        if assumptions:
-            multi_run_flag = isinstance(assumptions, dict)
-        else:
-            multi_run_flag = False 
-        
-        # overwrite any custom_endpoint
-        url = f"{self.url}/run_deal"
-        if custom_endpoint:
-            url = f"{self.url}/{custom_endpoint}"
+        multi_run_flag = True if isinstance(assumptions, dict) else False
+        url = f"{self.url}/runDealByScenarios"  if multi_run_flag  else f"{self.url}/runDeal"
 
         if isinstance(deal, str):
             with open(deal,'rb') as _f:
@@ -200,31 +211,56 @@ class API:
 
         result = self._send_req(req,url)
 
-        if read:
-            if multi_run_flag:
+        if read and multi_run_flag:
                 return { n:deal.read(_r,position=position) for (n,_r) in result.items()}
-            else:
+        elif read :
                 return deal.read(result,position=position)
         else:
             return result
 
-    def runPool(self, pool, assumptions=[],read=True):
-        url = f"{self.url}/run_pool"        
-        req = self.build_pool_req(pool, assumptions=assumptions)
-
-        result = self._send_req(req,url)
-
-        if read:
-            flow_header,idx = guess_pool_flow_header(result[0],self.lang)
+    def runPool(self, pool, assumptions,read=True):
+        def read_single(pool_resp):
+            flow_header,idx = guess_pool_flow_header(pool_resp[0],pool_lang)
             try:
-                result = pd.DataFrame([_['contents'] for _ in result] , columns=flow_header)
+                result = pd.DataFrame([_['contents'] for _ in pool_resp] , columns=flow_header)
             except ValueError as e:
                 logging.error(f"Failed to match header:{flow_header} with {result[0]['contents']}")
             result = result.set_index(idx)
             result.index.rename(idx, inplace=True)
             result.sort_index(inplace=True)
-        return result
+            return result
+            
+        multi_scenario = True if isinstance(assumptions, dict) else False 
+        url = f"{self.url}/runPoolByScenarios" if multi_scenario else f"{self.url}/runPool"
+        pool_lang = guess_pool_locale(pool)
+        req = self.build_pool_req(pool, assumptions=assumptions)
+        result = self._send_req(req, url)
+
+        if read and (not multi_scenario):
+            return read_single(result)
+        elif read and multi_scenario : 
+            return mapValsBy(result, read_single)
+        else:
+            return result
+
     
+    def runStructs(self, deals, assumptions=None, pricing=None, read=True ):
+        assert isinstance(deals, dict),f"Deals should be a dict but got {deals}"
+        url = f"{self.url}/runMultiDeals" 
+        _assumptions = mkAssumption2(assumptions) if assumptions else None 
+        _pricing = mkPricingAssump(pricing) if pricing else None
+        req = json.dumps(mkTag(("MultiDealRunReq"
+                                ,[{k:v.json for k,v in deals.items()}
+                                  ,_assumptions
+                                  ,_pricing]))
+                        ,ensure_ascii=False)
+
+        result = self._send_req(req,url)
+        if read :
+            return {k:deals[k].read(v) for k,v in result.items()}    
+        else:
+            return result
+
     def _send_req(self,_req,_url)->dict:
         try:
             r = requests.post(_url, data=_req.encode('utf-8'), headers=self.hdrs, verify=False)
@@ -240,7 +276,10 @@ class API:
         except JSONDecodeError as e:
             raise RuntimeError(e)        
 
+
+
 def guess_pool_flow_header(x,l):
+    assert isinstance(x, dict), f"x is not a map but {x}, type:{type(x)}"
     match (x['tag'],l):
         case ('MortgageFlow','chinese'):
             return (china_mortgage_flow_fields_d,"日期")
@@ -266,55 +305,3 @@ def save(deal,p:str):
     match deal:
         case _:
             save_to(deal)
-
-def init_jupyter():
-    pd.set_option('display.max_rows', None)
-    pd.set_option('display.max_columns', None)
-    pd.set_option('display.width', None)
-    pd.set_option('display.max_colwidth', None)
-    pd.options.display.float_format = '{:,}'.format
-
-
-def comp_df(x, y):
-    return pd.merge(x, y, on="日期", how='outer').sort_index().sort_index(axis=1)
-
-def comp_engines(engine1,engine2, d, a=None):
-    
-    r1 = engine1.run(d,assumptions=a,read=True)
-    r2 = engine2.run(d,assumptions=a,read=True)
-
-
-    comp_result = {}
-    # pool check
-    if not r1['pool']['flow'].equals(r2['pool']['flow']):
-        comp_result['pool'] = comp_df(r1['pool']['flow'],r2['pool']['flow'])
-    else:
-        comp_result['pool'] = True
-
-    # expense check  
-    comp_result['fee'] = {}
-    for fn,f in r1['fees'].items():
-        if not f.equals(r2['fees'][fn]):
-            comp_result['fee'][fn] = comp_df(f,r2['fees'][fn])
-        else:
-            comp_result['fee'][fn] = True
-
-    # bond check
-    comp_result['bond'] = {}
-    for fn,f in r1['bonds'].items():
-        if not f.equals(r2['bonds'][fn]):
-            comp_result['bond'][fn] = comp_df(f,r2['bonds'][fn])
-        else:
-            comp_result['bond'][fn] = True
-
-
-    # account check
-    comp_result['account'] = {}
-    for fn,f in r1['accounts'].items():
-        if not f.equals(r2['accounts'][fn]):
-            comp_result['account'][fn] = comp_df(f,r2['accounts'][fn])
-        else:
-            comp_result['account'][fn] = True
-
-    return comp_result
-
