@@ -2,7 +2,7 @@ from dataclasses import dataclass
 import functools
 
 from absbox import *
-from absbox.local.util import mkTag
+from absbox.local.util import mkTag,mapListValBy,mapValsBy,renameKs2
 from absbox.local.component import *
 from absbox.local.base import * 
 import pandas as pd
@@ -19,11 +19,12 @@ class Generic:
     fees: tuple
     waterfall: dict
     collection: list
-    call: tuple = None 
     liqFacility :dict = None
-    custom: dict = None 
+    rateSwap:dict = None
+    currencySwap:dict = None
     trigger:dict = None
     status:str = "Amortizing"
+    custom: dict = None
 
     @property
     def json(self):
@@ -31,6 +32,7 @@ class Generic:
         distsAs,collectsAs,cleansAs = [ [ mkWaterfall2(_action) for _action in _actions] for _actions in [dists,collects,cleans] ]
         distsflt,collectsflt,cleanflt = [ itertools.chain.from_iterable(x) for x in [distsAs,collectsAs,cleansAs] ]
         parsedDates = mkDate(self.dates)
+        closingDate = self.dates['closing']
         """
         get the json formatted string
         """
@@ -42,36 +44,19 @@ class Generic:
                      , "asOfDate": self.dates['cutoff']
                      , "issuanceStat": self.pool.get("issuanceStat")
                      , "futureCf":mkCf(self.pool.get('cashflow', [])) },
-            "bonds": functools.reduce(lambda result, current: result | current
-                                      , [mk(['bond', bn, bo]) for (bn, bo) in self.bonds]),
+            "bonds": { bn: mkBnd(bn,bo) for (bn,bo) in self.bonds},
             "waterfall": mkWaterfall({},self.waterfall.copy()),  
-            "fees": functools.reduce(lambda result, current: result | current
-                                     , [mk(["fee", feeName, feeO]) for (feeName, feeO) in self.fees]) if self.fees else {},
-            "accounts": functools.reduce(lambda result, current: result | current
-                                         , [mk(["account", accName, accO]) for (accName, accO) in self.accounts]),
-            "collects": self.collection
+            "fees": {fn: mkFee(fo|{"name":fn},fsDate = closingDate) 
+                                 for (fn,fo) in self.fees},
+            "accounts": {an:mkAcc(an,ao) for (an,ao) in self.accounts},
+            "collects": [ mkCollection(c) for c in self.collection],
+            "rateSwap": { k:mkRateSwap(v) for k,v in self.currencySwap.items()} if self.rateSwap else None,
+            "currencySwap":None ,
+            "custom": {cn:mkCustom(co) for cn,co in self.custom.items()} if self.custom else None ,
+            "triggers": renameKs2(mapListValBy(self.trigger,mkTrigger),englishDealCycle) if self.trigger else None,
+            "liqProvider": {ln: mkLiqProvider(ln, lo | {"start":closingDate} ) 
+                               for ln,lo in self.liqFacility.items() } if self.liqFacility else None
         }
-        for fn, fo in _r['fees'].items():
-            if fo['feeStart'] is None:
-                fo['feeStart'] = self.dates['closing']
-
-        if hasattr(self, "custom") and self.custom is not None:
-            _r["custom"] = {}
-            for n,ci in self.custom.items():
-                _r["custom"][n] = mkCustom(ci)
-        
-        if hasattr(self, "trigger") and self.trigger is not None:
-            _trigger  = self.trigger
-            _trr = {mkWhenTrigger(_loc):
-                       [[mkTrigger(_trig),mkTriggerEffect(_effect)] for (_trig,_effect) in _vs ] 
-                       for _loc,_vs in _trigger.items()}
-            _r["triggers"] = _trr
-        
-        if hasattr(self, "liqFacility") and self.liqFacility is not None:
-            _providers = {}
-            for (_k, _p) in self.liqFacility.items():
-                _providers[_k] = mkLiqProvider(_k, ( _p | {"start": self.dates['closing']}))
-            _r["liqProvider"] = _providers
 
         _dealType = identify_deal_type(_r)
 
@@ -90,11 +75,17 @@ class Generic:
     def read(self, resp, position=None):
         read_paths = {'bonds': ('bndStmt' , english_bondflow_fields , "bond")
                      , 'fees': ('feeStmt' , english_fee_flow_fields_d , "fee")
-                     , 'accounts': ('accStmt' , english_acc_flow_fields_d , "account")}
+                     , 'accounts': ('accStmt' , english_acc_flow_fields_d , "account")
+                     , 'liqProvider': ('liqStmt', english_liq_flow_fields_d, "")
+                     , 'rateSwap': ('rsStmt', english_rs_flow_fields_d, "")
+                     }
+        deal_content = resp[0]['contents']
         output = {}
         for comp_name, comp_v in read_paths.items():
+            if deal_content[comp_name] is None:
+                continue
             output[comp_name] = {}
-            for k, x in resp[0][comp_name].items():
+            for k, x in deal_content[comp_name].items():
                 ir = None
                 if x[comp_v[0]]:
                     ir = [_['contents'] for _ in x[comp_v[0]]]
@@ -105,27 +96,11 @@ class Generic:
                           for f, v in output['fees'].items()}
 
         # aggregate accounts
-        agg_acc = {}
-        for k,v in output['accounts'].items():
-            acc_by_date = v.groupby("date")
-            acc_txn_amt = acc_by_date.agg(change=("change", sum))
-            ending_bal_column = acc_by_date.last()['balance'].rename("end balance")
-            begin_bal_column = ending_bal_column.shift(1).rename("begin balance")
-            agg_acc[k] = acc_txn_amt.join([begin_bal_column,ending_bal_column])
-            if agg_acc[k].empty:
-                agg_acc[k].columns = ['begin balance', "change", 'end balance']
-                continue
-            fst_idx = agg_acc[k].index[0]
-            agg_acc[k].at[fst_idx, 'begin balance'] = round(agg_acc[k].at[fst_idx, 'end balance'] - agg_acc[k].at[fst_idx, 'change'], 2)
-            agg_acc[k] = agg_acc[k][['begin balance', "change", 'end balance']]
-
-        output['agg_accounts'] = agg_acc
+        output['agg_accounts'] = aggAccs(output['accounts'], 'en')
 
         output['pool'] = {}
-        
-        _pool_cf_header,_ = guess_pool_flow_header(resp[0]['pool']['futureCf'][0],"english")
-        
-        output['pool']['flow'] = pd.DataFrame([_['contents'] for _ in resp[0]['pool']['futureCf']]
+        _pool_cf_header,_ = guess_pool_flow_header(deal_content['pool']['futureCf'][0],"english")
+        output['pool']['flow'] = pd.DataFrame([_['contents'] for _ in deal_content['pool']['futureCf']]
                                               , columns=_pool_cf_header)
         pool_idx = 'Date'
         output['pool']['flow'] = output['pool']['flow'].set_index(pool_idx)
